"""
    Implements a full text search engine.
"""

import os
from datetime import datetime
from typing import List

import pytz  # type: ignore
from whoosh import fields  # type: ignore
from whoosh.analysis import FancyAnalyzer  # type: ignore
from whoosh.compat import u  # type: ignore
from whoosh.filedb.filestore import FileStorage  # type: ignore
from whoosh.qparser import QueryParser  # type: ignore
from whoosh.qparser.dateparse import DateParserPlugin  # type: ignore

from vids_db.models import Video

SCHEMA = fields.Schema(
    url=fields.ID(stored=True, unique=True, sortable=True),
    channel_name=fields.TEXT(
        stored=True, chars=True, vector=True, analyzer=FancyAnalyzer()
    ),
    date=fields.DATETIME(stored=True, sortable=True),
    # title=fields.TEXT(
    #    stored=True, chars=True, vector=True, analyzer=FancyAnalyzer()
    # ),
    title=fields.NGRAMWORDS(stored=True, minsize=3),
    views=fields.NUMERIC(stored=True, sortable=True, bits=64),
)


def _filter_out_duplicate_videos(videos: List[Video]) -> List[Video]:
    found_urls = set()
    filtered_videos = []
    for video in videos:
        if video.url in found_urls:
            continue
        found_urls.add(video.url)
        filtered_videos.append(video)
    if len(filtered_videos) != len(videos):
        print(
            f"Filtered out {len(videos) - len(filtered_videos)} duplicate videos."
        )
    return filtered_videos


class DbFullTextSearch:
    """Impelmentation of a full text search database."""

    def __init__(self, index_path) -> None:
        """Initialize the database."""
        self.storage = FileStorage(index_path)
        if self.storage.index_exists():
            self.index = self.storage.open_index()
        else:
            os.makedirs(index_path, exist_ok=True)
            self.index = self.storage.create_index(SCHEMA)

    def clear(self) -> None:  # pylint: disable=no-self-use
        """Clear the database."""
        # TODO: Implement this.  # pylint: disable=W0511
        return

    def add_videos(self, videos: List[Video]) -> None:
        """Add videos to the database."""
        videos = _filter_out_duplicate_videos(videos)
        with self.index.writer() as writer:
            with writer.group():
                for vid in videos:
                    published: datetime = vid.date_published
                    # Change published datetime to utc timezone.
                    published_utc = published.astimezone(pytz.utc)
                    writer.update_document(
                        url=vid.url,
                        channel_name=u(vid.channel_name),
                        date=published_utc,
                        title=u(vid.title),
                        views=vid.views,
                    )

    def _field_search(
        self, field_name: str, query_string: str, limit: int = 40
    ) -> List[dict]:
        """Searcher for videos by one of the fields."""
        qparser = QueryParser(field_name, schema=SCHEMA)
        qparser.add_plugin(DateParserPlugin(free=False))
        qry = qparser.parse(query_string)
        with self.index.searcher() as searcher:
            # matcher = query.matcher(searcher)  # useful for debugging
            results = searcher.search(qry, mask=None, limit=limit, terms=True)
            # Convert the results to dicts.
            results_dicts = []
            for result in results:
                results_dicts.append(
                    {
                        "url": result["url"],
                        "channel_name": result["channel_name"],
                        "date": result["date"],
                        "title": result["title"],
                        "views": result["views"],
                    }
                )
            return results_dicts

    def title_search(self, query_string: str, limit: int = 40) -> List[dict]:
        """Searcher for videos by title."""
        return self._field_search("title", query_string, limit)

    def channel_search(self, query_string: str, limit: int = 40) -> List[dict]:
        """Searcher for videos by title."""
        return self._field_search("channel_name", query_string, limit)



# pylint: disable=all
import os
from datetime import datetime
from typing import List, Optional

from vids_db.db_full_text_search import DbFullTextSearch
from vids_db.db_sqlite_video import DbSqliteVideo  # type: ignore
from vids_db.models import Video

HERE = os.path.dirname(os.path.abspath(__file__))
PROJECT_ROOT = os.path.dirname(HERE)
DB_PATH_DIR = os.path.join(PROJECT_ROOT, "data")

FULL_TEXT_SEARCH_ENABLED = (
    os.environ.get("FULL_TEXT_SEARCH_ENABLED", "0") == "1"
)


class Database:
    def __init__(self, db_path: Optional[str] = None) -> None:
        db_path = db_path or DB_PATH_DIR
        os.makedirs(db_path, exist_ok=True)
        self.db_path = db_path
        db_path_sqlite = os.path.join(db_path, "videos.sqlite")
        self.db_full_text_search = None
        full_text_enabled = (
            os.environ.get("FULL_TEXT_SEARCH_ENABLED", "0") == "1"
        )
        if full_text_enabled:
            db_path_fts = os.path.join(db_path, "full_text_seach")
            self.db_full_text_search = DbFullTextSearch(db_path_fts)
        self.db_sqlite = DbSqliteVideo(db_path_sqlite)

    def clear(self) -> None:
        self.db_sqlite.clear()
        if self.db_full_text_search:
            self.db_full_text_search.clear()

    def update_many(self, vids: List[Video]) -> None:  # type: ignore
        self.db_sqlite.insert_or_update(vids)
        if self.db_full_text_search:
            self.db_full_text_search.add_videos(vids)

    def update(self, vid: Video) -> None:
        self.update_many([vid])

    def get_channel_names(self) -> List[str]:
        return self.db_sqlite.get_channel_names()

    def get_by_urls(self, urls: List[str]) -> List[Video]:
        return self.db_sqlite.find_videos_by_urls(urls)

    def remove_by_channel_name(self, channel_name: str) -> None:
        self.db_sqlite.remove_by_channel_name(channel_name)

    def get_video_list(
        self,
        date_start: datetime,
        date_end: datetime,
        channel_name: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> List[Video]:
        vid_list: List[Video] = self.db_sqlite.find_videos(
            date_start, date_end, channel_name=channel_name, limit_count=limit
        )
        return vid_list

    def query_video_list(
        self,
        query_string: str,
        limit: Optional[int] = None,
    ) -> List[Video]:
        if not self.db_full_text_search:
            return []
        options = {}
        if limit is not None:
            options["limit"] = limit
        vids0: List[dict] = self.db_full_text_search.channel_search(
            query_string, **options
        )
        vids1: List[dict] = self.db_full_text_search.title_search(
            query_string, **options
        )
        vid_list = vids0 + vids1
        found_urls = set()
        filtered_vids = []
        for vid in vid_list:
            if vid["url"] in found_urls:
                continue
            found_urls.add(vid["url"])
            filtered_vids.append(vid)
        if limit is not None:
            filtered_vids = filtered_vids[:limit]

        urls = [v["url"] for v in filtered_vids]
        vids = self.db_sqlite.find_videos_by_urls(urls)
        return vids



# pylint: disable=all
# types: disable=all

from datetime import datetime
from typing import Union

import pytz  # type: ignore
from dateutil.parser import parse


def _my_date_parse(date_string: str) -> datetime:
    try:
        return datetime.fromisoformat(date_string)
    except ValueError as verr:
        if "Invalid isoformat" in str(verr):
            return parse(date_string, fuzzy=True)
        else:
            raise


def parse_datetime(s, tzinfo=None) -> datetime:  # type: ignore
    date: datetime = _my_date_parse(s)
    if tzinfo:
        if isinstance(tzinfo, str):
            tzinfo = pytz.timezone(tzinfo)
        if date.tzinfo is None:
            date = date.replace(tzinfo=tzinfo)
        else:
            date = date.astimezone(tzinfo)
    return date


def iso_fmt(date_obj: Union[datetime, str]) -> str:
    if isinstance(date_obj, datetime):
        return date_obj.isoformat()
    elif isinstance(date_obj, str):
        return parse_datetime(date_obj).isoformat()
    else:
        raise ValueError(f"{__file__} Unexpected type: {type(date_obj)}")



# disable pylint for the entire file
# pylint: disable=all

from __future__ import annotations

import json
from datetime import datetime, timedelta
from typing import Dict, List, Optional, Union

from pydantic import (
    AnyUrl,
    BaseModel,
    NonNegativeFloat,
    NonNegativeInt,
    constr,
    validator,
)

from vids_db.date import iso_fmt, parse_datetime


def parse_duration(duration: str) -> float:
    """
    Checks that the duration is in the format HH:MM:SS.
    Other acceptable formats include SS.
    Ok:
      ""
      "?"
      06
      6
      60
      61
      23:24
      23:24:01.34
    Not Ok:
      -7
      61  # above 60 seconds
      61:01 # above 60 minutes
      25:24:01.34 # above 24 hours
    """

    def _raise() -> None:
        raise ValueError(f"Invalid duration: {duration}")

    def _is_non_neg_int(s: str) -> bool:
        try:
            return int(s) >= 0
        except ValueError:
            return False

    def _is_non_neg_float(s: str) -> bool:
        try:
            return float(s) >= 0.0
        except ValueError:
            return False

    try:
        valf = float(duration)
        if valf >= 0.0:
            return valf
        else:
            _raise()
    except ValueError:
        pass

    if "" == duration or "?" == duration or "Live" == duration:
        return 0
    # Simple case
    if ":" not in duration and "." not in duration:
        try:
            tmp = float(duration)
            if tmp < 0:
                _raise()
            return tmp
        except ValueError:
            _raise()
    new_duration = duration
    units = new_duration.split(":")
    if len(units) > 3 or len(units) < 1:
        _raise()
    units.reverse()
    total: float = 0.0
    limit_multiplier = [
        (60, 1),
        (60, 60),
        (24, 60 * 24),
    ]
    for i, unit in enumerate(units):
        if i == 0:
            is_valid_number = _is_non_neg_float(unit)
        else:
            is_valid_number = _is_non_neg_int(unit)
        if not is_valid_number:
            _raise()
        limit, multipler = limit_multiplier[i]
        val = float(unit)
        if val >= limit:
            _raise()
        total += val * multipler
    return total


class Video(BaseModel):
    """Represents a video object."""

    channel_name: constr(min_length=2)  # type: ignore
    title: constr(min_length=2)  # type: ignore
    date_published: datetime  # from the scraped website
    date_lastupdated: datetime
    channel_url: AnyUrl
    source: constr(min_length=4)  # type: ignore
    url: AnyUrl
    duration: NonNegativeFloat
    description: str
    img_src: AnyUrl
    iframe_src: str
    views: NonNegativeInt
    # rank: Optional[float] = None  # optional stdev rank.

    @validator("duration", pre=True)
    def check_duration(cls, v):
        return parse_duration(v)

    @validator("date_published", pre=True)
    def check_date_published(cls, v):
        return iso_fmt(v)

    @validator("date_lastupdated", pre=True)
    def check_date_lastupdated(cls, v):
        return iso_fmt(v)

    @validator("views", pre=True)
    def check_views(cls, v):
        if v == "" or v == "?":
            return 0
        try:
            return int(v)
        except ValueError:
            return 0

    @classmethod
    def from_list_of_dicts(cls, data: List[Dict]) -> List[Video]:
        out: List[Video] = []
        for datum in data:
            vid = Video(**datum)
            out.append(vid)
        return out

    @classmethod
    def to_plain_list(cls, data: List[Video]) -> List[Dict]:
        out = []
        vid: Video
        for vid in data:
            out.append(vid.dict())
        return out

    @classmethod
    def parse_json(cls, data: Union[str, dict]) -> List[dict]:
        """
        Parses a string or json dict and returns a json dict representation
        that can be used in a network request.
        """
        out: List[dict] = []
        if isinstance(data, str):
            json_data = json.loads(data)
        else:
            json_data = data
        if "content" in json_data:  # This is the publishing format.
            json_data = json_data["content"]
        for json_video in json_data:
            try:
                vid = Video(**json_video)
                out.append(vid.to_json())
            except Exception as err:
                print(
                    f"{__file__}: Skipping {json_video.get('url')} because {err}"
                )
        return out

    def video_age_seconds(self, now_time: Optional[datetime] = None) -> float:
        """
        Returns the date published as a datetime object.
        """
        now_time = now_time or datetime.now()
        diff: timedelta = now_time - parse_datetime(self.date_published)
        return diff.total_seconds()

    def to_json(self) -> dict:
        """
        Returns a json representation of the video object.
        """
        data = self.dict()
        data["date_published"] = self.date_published.isoformat()
        data["date_lastupdated"] = self.date_lastupdated.isoformat()
        return data

    def to_json_str(self) -> str:
        """
        Returns a json string representation of the video object.
        """
        return json.dumps(self.to_json(), ensure_ascii=False)
